###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""This file contains code to handle InvalidPipeline exceptions that contain
upgrade requests."""
from __future__ import division

from vistrails.core import debug
import vistrails.core.db.action
from vistrails.core.modules.module_registry import get_module_registry, \
     ModuleDescriptor, MissingModule, MissingPort, MissingPackage
from vistrails.core.modules.utils import parse_descriptor_string, \
    create_descriptor_string, parse_port_spec_string, create_port_spec_string
from vistrails.core.packagemanager import get_package_manager
from vistrails.core.system import get_vistrails_basic_pkg_id
from vistrails.core.vistrail.annotation import Annotation
from vistrails.core.vistrail.module_control_param import ModuleControlParam
from vistrails.core.vistrail.connection import Connection
from vistrails.core.vistrail.port import Port
from vistrails.core.vistrail.port_spec import PortSpec
from vistrails.core.vistrail.port_spec_item import PortSpecItem
from vistrails.core.utils import versions_increasing
import copy

##############################################################################

class UpgradeWorkflowError(Exception):

    def __init__(self, msg, module=None, port_name=None, port_type=None):
        Exception.__init__(self, msg)
        self._msg = msg
        self._module = module
        self._port_name = port_name
        self._port_type = port_type.lower() if port_type else None
        
    def __str__(self):
        return "Upgrading workflow failed.\n" + self._msg

class UpgradeModuleRemap(object):
    def __init__(self, start_version, end_version, 
                 output_version, new_module=None,
                 dst_port_remap=None, src_port_remap=None,
                 function_remap=None, annotation_remap=None,
                 control_param_remap=None, module_name=None):
        self.module_name = module_name
        self.start_version = start_version
        self.end_version = end_version
        self.output_version = output_version
        self.new_module = new_module

        if dst_port_remap is None:
            self._dst_port_remap = {}
        else:
            self._dst_port_remap = dst_port_remap
        if src_port_remap is None:
            self._src_port_remap = {}
        else:
            self._src_port_remap = src_port_remap
        if function_remap is None:
            self._function_remap = {}
        else:
            self._function_remap = function_remap
        if annotation_remap is None:
            self._annotation_remap = {}
        else:
            self._annotation_remap = annotation_remap
        if control_param_remap is None:
            self._control_param_remap = {}
        else:
            self._control_param_remap = control_param_remap

    @classmethod
    def __copy__(cls, obj):
        newobj = cls()
        for k, v in obj.__dict__.iteritems():
            if k.startswith('_') and k.endswith('_remap'):
                v = copy.copy(v)
            newobj.__dict__[k] = v
        return newobj

    @classmethod
    def from_tuple(cls, module_name, t):
        if len(t) == 3:
            obj = cls(t[0], t[1], None, t[2], module_name=module_name)
            remap = None
        elif len(t) == 4:
            obj = cls(t[0], t[1], None, t[2], module_name=module_name)
            remap = t[3]
        elif len(t) == 5:
            obj = cls(t[0], t[1], t[2], t[3], module_name=module_name)
            remap = t[4]
        else:
            raise TypeError("UpgradeModuleRemap.from_tuple() got a tuple of "
                            "length %d" % len(t))
        if remap is not None:
            for remap_type, remap_dict in remap.iteritems():
                for remap_name, remap_change in remap_dict.iteritems():
                    obj.add_remap(remap_type, remap_name, remap_change)
        return obj

    def _get_dst_port_remap(self):
        return self._dst_port_remap
    dst_port_remap = property(_get_dst_port_remap)

    def _get_src_port_remap(self):
        return self._src_port_remap
    src_port_remap = property(_get_src_port_remap)
    
    def _get_function_remap(self):
        # !!! we're going to let dst_port_remap serve as a
        # base for function_remap but the developer is
        # responsible for knowing that anything beyond name
        # remaps requires different functions
        d = copy.copy(self._dst_port_remap)
        d.update(self._function_remap)
        return d
    function_remap = property(_get_function_remap)
    
    def _get_annotation_remap(self):
        return self._annotation_remap
    annotation_remap = property(_get_annotation_remap)    

    def _get_control_param_remap(self):
        return self._control_param_remap
    control_param_remap = property(_get_control_param_remap)    

    def add_remap(self, remap_type, remap_name, remap_change):
        if not hasattr(self, '_%s' % remap_type):
            raise ValueError('remap_type "%s" not allowed' % remap_type)
        d = getattr(self, '_%s' % remap_type)
        d[remap_name] = remap_change
            
        # if remap_type not in self.allowed_remaps:
        #     raise ValueError("remap_type must be one of %s" % allowed_remaps)
        # self.remap[remap_type][remap_name] = remap_change
        
    def get_remap(self, remap_type):
        if not hasattr(self, '_%s' % remap_type):
            raise ValueError('remap_type "%s" not allowed' % remap_type)
        d = getattr(self, '_%s' % remap_type)
        return d

        # if remap_type not in self.allowed_remaps:
        #     raise ValueError("remap_type must be one of %s" % allowed_remaps)
        # return self.remap[remap_type]

    def get_output_version(self):
        return self.output_version

class UpgradePackageRemap(object):
    def __init__(self):
        self.remaps = {}  # name (str): remap (UpgradeModuleRemap)

    @classmethod
    def __copy__(cls, obj):
        newobj = cls()
        newobj.remaps = dict((modname, copy.copy(modremap))
                             for modname, modremap in obj.remaps.iteritems())
        return newobj

    @classmethod
    def from_dict(cls, d):
        pkg_remap = cls()
        for module_name, remap_list in d.iteritems():
            for remap in remap_list:
                pkg_remap.add_module_remap(remap, module_name)
        return pkg_remap

    def add_module_remap(self, module_remap, module_name=None):
        if isinstance(module_remap, tuple):
            if module_name is None:
                raise ValueError("module_name must be specified if "
                                 "module_remap is a tuple")
            module_remap = UpgradeModuleRemap.from_tuple(module_name, 
                                                         module_remap)
        else:
            if module_name is not None:
                # assume user wants to override name
                module_remap.module_name = module_name
        if module_remap.module_name not in self.remaps:
            self.remaps[module_remap.module_name] = []
        self.remaps[module_remap.module_name].append(module_remap)

    def get_module_remaps(self, module_name):
        if module_name in self.remaps:
            return self.remaps[module_name]
        return []

    def has_module_remaps(self, module_name):
        return module_name in self.remaps

    def get_module_upgrade(self, module_name, old_version):
        for module_remap in self.get_module_remaps(module_name):
            if ((module_remap.start_version is None or 
                 not versions_increasing(old_version, 
                                         module_remap.start_version)) and
                (module_remap.end_version is None or
                 versions_increasing(old_version, 
                                     module_remap.end_version))):
                return module_remap
        return None

class UpgradeWorkflowHandler(object):

    @staticmethod
    def dispatch_request(controller, module_id, current_pipeline):
        pm = get_package_manager()
        if module_id not in current_pipeline.modules:
            # It is possible that some other upgrade request has
            # already removed the invalid module of this request. In
            # that case, disregard the request.
            debug.log("module %s already handled. skipping" % module_id)
            return []
        invalid_module = current_pipeline.modules[module_id]
        pkg = pm.get_package(invalid_module.package)
        if hasattr(pkg.module, 'handle_module_upgrade_request'):
            f = pkg.module.handle_module_upgrade_request
            return f(controller, module_id, current_pipeline)
        elif hasattr(pkg.module, '_upgrades'):
            return UpgradeWorkflowHandler.remap_module(controller, module_id, 
                                                       current_pipeline,
                                                       pkg.module._upgrades)
        else:
            debug.log('Package "%s" cannot handle upgrade request. '
                      'VisTrails will attempt automatic upgrade.' % \
                          pkg.identifier)
            auto_upgrade = UpgradeWorkflowHandler.attempt_automatic_upgrade
            return auto_upgrade(controller, current_pipeline, module_id)

    @staticmethod
    def check_port_spec(module, port_name, port_type, descriptor=None, 
                        sigstring=None):
        basic_pkg = get_vistrails_basic_pkg_id()

        reg = get_module_registry()
        found = False
        try:
            if descriptor is not None:
                s = reg.get_port_spec_from_descriptor(descriptor, port_name,
                                                      port_type)
                found = True

                spec_tuples = parse_port_spec_string(sigstring, basic_pkg)
                for i in xrange(len(spec_tuples)):
                    spec_tuple = spec_tuples[i]
                    port_pkg = reg.get_package_by_name(spec_tuple[0])
                    if port_pkg.identifier != spec_tuple[0]:
                        # we have an old identifier
                        spec_tuples[i] = (port_pkg.identifier,) + spec_tuple[1:]
                sigstring = create_port_spec_string(spec_tuples)
                # sigstring = expand_port_spec_string(sigstring, basic_pkg)
                if s.sigstring != sigstring:
                    msg = ('%s port "%s" of module "%s" exists, but '
                           'signatures differ "%s" != "%s"') % \
                           (port_type.capitalize(), port_name, module.name,
                            s.sigstring, sigstring)
                    raise UpgradeWorkflowError(msg, module, port_name, port_type)
        except MissingPort:
            pass

        if not found and \
                not module.has_portSpec_with_name((port_name, port_type)):
            msg = '%s port "%s" of module "%s" does not exist.' % \
                (port_type.capitalize(), port_name, module.name)
            raise UpgradeWorkflowError(msg, module, port_name, port_type)

    @staticmethod
    def find_descriptor(controller, pipeline, module_id, desired_version=''):
        reg = get_module_registry()

        get_descriptor = reg.get_descriptor_by_name
        pm = get_package_manager()
        invalid_module = pipeline.modules[module_id]
        mpkg, mname, mnamespace, mid = (invalid_module.package,
                                        invalid_module.name,
                                        invalid_module.namespace,
                                        invalid_module.id)
        pkg = pm.get_package(mpkg)
        desired_version = ''
        d = None
        # don't check for abstraction/subworkflow since the old module
        # could be a subworkflow
        if reg.has_abs_upgrade(*invalid_module.descriptor_info):
            return reg.get_abs_upgrade(*invalid_module.descriptor_info)

        try:
            try:
                d = get_descriptor(mpkg, mname, mnamespace, '', desired_version)
            except MissingModule, e:
                r = None
                if pkg.can_handle_missing_modules():
                    r = pkg.handle_missing_module(controller, module_id, 
                                                  pipeline)
                    d = get_descriptor(mpkg, mname, mnamespace, '', 
                                       desired_version)
                if not r:
                    raise e
        except MissingModule, e:
            return None
        assert isinstance(d, ModuleDescriptor)
        return d

    @staticmethod
    def check_upgrade(pipeline, module_id, d, function_remap=None,
                      src_port_remap=None, dst_port_remap=None):
        if function_remap is None:
            function_remap = {}
        if src_port_remap is None:
            src_port_remap = {}
        if dst_port_remap is None:
            dst_port_remap = {}
        invalid_module = pipeline.modules[module_id]
        def check_connection_port(port):
            port_type = PortSpec.port_type_map.inverse[port.type]
            UpgradeWorkflowHandler.check_port_spec(invalid_module,
                                                   port.name, port_type,
                                                   d, port.sigstring)
            
        # check if connections are still valid
        for _, conn_id in pipeline.graph.edges_from(module_id):
            port = pipeline.connections[conn_id].source
            if port.name not in src_port_remap:
                check_connection_port(port)
        for _, conn_id in pipeline.graph.edges_to(module_id):
            port = pipeline.connections[conn_id].destination
            if port.name not in dst_port_remap:
                check_connection_port(port)

        # check if function values are still valid
        for function in invalid_module.functions:
            # function_spec = function.get_spec('input')
            if function.name not in function_remap:
                UpgradeWorkflowHandler.check_port_spec(invalid_module,
                                                       function.name, 
                                                       'input', d,
                                                       function.sigstring)

    @staticmethod
    def attempt_automatic_upgrade(controller, pipeline, module_id,
                                  function_remap=None, src_port_remap=None, 
                                  dst_port_remap=None, annotation_remap=None,
                                  control_param_remap=None):
        """attempt_automatic_upgrade(module_id, pipeline): [Action]

        Attempts to automatically upgrade module by simply adding a
        new module with the current package version, and recreating
        all connections and functions. If any of the ports used are
        not available, raise an exception that will trigger the
        failure of the entire upgrade.

        attempt_automatic_upgrade returns a list of actions if 
        successful.
        """

        invalid_module = pipeline.modules[module_id]
        mpkg, mname, mnamespace, mid = (invalid_module.package,
                                        invalid_module.name,
                                        invalid_module.namespace,
                                        invalid_module.id)
        d = UpgradeWorkflowHandler.find_descriptor(controller, pipeline, 
                                                   module_id)
        if not d:
            if mnamespace:
                nss = mnamespace + '|' + mname
            else:
                nss = mname
            msg = ("Could not upgrade module %s from package %s.\n" %
                    (nss, mpkg))
            raise UpgradeWorkflowError(msg)

        UpgradeWorkflowHandler.check_upgrade(pipeline, module_id, d, 
                                             function_remap, 
                                             src_port_remap, dst_port_remap)

        # If we passed all of these checks, then we consider module to
        # be automatically upgradeable. Now create actions that will
        # delete functions, module, and connections, and add new
        # module with corresponding functions and connections.

        return UpgradeWorkflowHandler.replace_module(controller, pipeline, 
                                                     module_id, d, 
                                                     function_remap,
                                                     src_port_remap, 
                                                     dst_port_remap,
                                                     annotation_remap,
                                                     control_param_remap)

    @staticmethod
    def create_new_connection(controller, src_module, src_port, 
                              dst_module, dst_port):
        # spec -> name, type, signature
        output_port_id = controller.id_scope.getNewId(Port.vtType)
        if isinstance(src_port, basestring):
            output_port_spec = src_module.get_port_spec(src_port, 'output')
            output_port = Port(id=output_port_id,
                               spec=output_port_spec,
                               moduleId=src_module.id,
                               moduleName=src_module.name)
        else:
            output_port = Port(id=output_port_id,
                               name=src_port.name,
                               type=src_port.type,
                               signature=src_port.signature,
                               moduleId=src_module.id,
                               moduleName=src_module.name)

        input_port_id = controller.id_scope.getNewId(Port.vtType)
        if isinstance(dst_port, basestring):
            input_port_spec = dst_module.get_port_spec(dst_port, 'input')
            input_port = Port(id=input_port_id,
                              spec=input_port_spec,
                              moduleId=dst_module.id,
                              moduleName=dst_module.name)
        else:
            input_port = Port(id=input_port_id,
                              name=dst_port.name,
                              type=dst_port.type,
                              signature=dst_port.signature,
                              moduleId=dst_module.id,
                              moduleName=dst_module.name)
        conn_id = controller.id_scope.getNewId(Connection.vtType)
        connection = Connection(id=conn_id,
                                ports=[input_port, output_port])
        return connection



    @staticmethod
    def replace_generic(controller, pipeline, old_module, new_module,
                        function_remap=None, src_port_remap=None, 
                        dst_port_remap=None, annotation_remap=None,
                        control_param_remap=None, use_registry=True):
        if function_remap is None:
            function_remap = {}
        if src_port_remap is None:
            src_port_remap = {}
        if dst_port_remap is None:
            dst_port_remap = {}
        if annotation_remap is None:
            annotation_remap = {}
        if control_param_remap is None:
            control_param_remap = {}

        basic_pkg = get_vistrails_basic_pkg_id()

        ops = []
        ops.extend(controller.delete_module_list_ops(pipeline, [old_module.id]))
        
        for annotation in old_module.annotations:
            if annotation.key not in annotation_remap:
                annotation_key = annotation.key
            else:
                remap = annotation_remap[annotation.key]
                if remap is None:
                    # don't add the annotation back in
                    continue
                elif not isinstance(remap, basestring):
                    ops.extend(remap(annotation))
                    continue
                else:
                    annotation_key = remap

            new_annotation = \
                Annotation(id=controller.id_scope.getNewId(Annotation.vtType),
                           key=annotation_key,
                           value=annotation.value)
            new_module.add_annotation(new_annotation)

        for control_param in old_module.control_parameters:
            if control_param.name not in control_param_remap:
                control_param_name = control_param.name
            else:
                remap = control_param_remap[control_param.name]
                if remap is None:
                    # don't add the control param back in
                    continue
                elif not isinstance(remap, basestring):
                    ops.extend(remap(control_param))
                    continue
                else:
                    control_param_name = remap

            new_control_param = \
                ModuleControlParam(id=controller.id_scope.getNewId(
                                                   ModuleControlParam.vtType),
                                   name=control_param_name,
                                   value=control_param.value)
            new_module.add_control_parameter(new_control_param)

        if not old_module.is_group() and not old_module.is_abstraction():
            for port_spec in old_module.port_spec_list:
                if port_spec.type == 'input':
                    if port_spec.name not in dst_port_remap:
                        spec_name = port_spec.name
                    else:
                        remap = dst_port_remap[port_spec.name]
                        if remap is None:
                            continue
                        elif not isinstance(remap, basestring):
                            ops.extend(remap(port_spec))
                            continue
                        else:
                            spec_name = remap
                elif port_spec.type == 'output':
                    if port_spec.name not in src_port_remap:
                        spec_name = port_spec.name
                    else:
                        remap = src_port_remap[port_spec.name]
                        if remap is None:
                            continue
                        elif not isinstance(remap, basestring):
                            ops.extend(remap(port_spec))
                            continue
                        else:
                            spec_name = remap                
                new_spec = port_spec.do_copy(True, controller.id_scope, {})
                new_spec.name = spec_name
                new_module.add_port_spec(new_spec)

        function_ops = []
        for function in old_module.functions:
            if function.name not in function_remap:
                function_name = function.name
            else:
                remap = function_remap[function.name]
                if remap is None:
                    # don't add the function back in
                    continue                    
                elif not isinstance(remap, basestring):
                    function_ops.extend(remap(function, new_module))
                    continue
                else:
                    function_name = remap

            if len(function.parameters) > 0:
                new_param_vals, aliases = zip(*[(p.strValue, p.alias) 
                                                for p in function.parameters])
            else:
                new_param_vals = []
                aliases = []
            if use_registry:
                function_port_spec = function_name
            else:
                def mk_psi(pos):
                    psi = PortSpecItem(module="Module", package=basic_pkg,
                                       namespace="", pos=pos)
                    return psi
                n_items = len(new_param_vals)
                function_port_spec = PortSpec(name=function_name,
                                              items=[mk_psi(i) 
                                                     for i in xrange(n_items)])
            new_function = controller.create_function(new_module, 
                                                      function_port_spec,
                                                      new_param_vals,
                                                      aliases)
            new_module.add_function(new_function)

        if None in function_remap:
            # used to add new functions
            remap = function_remap[None]
            function_ops.extend(remap(None, new_module))

        # add the new module
        ops.append(('add', new_module))
        ops.extend(function_ops)

        create_new_connection = UpgradeWorkflowHandler.create_new_connection

        for _, conn_id in pipeline.graph.edges_from(old_module.id):
            old_conn = pipeline.connections[conn_id]
            if old_conn.source.name not in src_port_remap:
                source_name = old_conn.source.name
            else:
                remap = src_port_remap[old_conn.source.name]
                if remap is None:
                    # don't add this connection back in
                    continue
                elif not isinstance(remap, basestring):
                    ops.extend(remap(old_conn, new_module))
                    continue
                else:
                    source_name = remap
                    
            old_dst_module = pipeline.modules[old_conn.destination.moduleId]
            if use_registry:
                source_port = source_name
            else:
                source_port = Port(name=source_name,
                                   type='source',
                                   signature=create_port_spec_string(
                                       [(basic_pkg, 'Variant', '')]))

            new_conn = create_new_connection(controller,
                                             new_module,
                                             source_port,
                                             old_dst_module,
                                             old_conn.destination)
            ops.append(('add', new_conn))
            
        for _, conn_id in pipeline.graph.edges_to(old_module.id):
            old_conn = pipeline.connections[conn_id]
            if old_conn.destination.name not in dst_port_remap:
                destination_name = old_conn.destination.name
            else:
                remap = dst_port_remap[old_conn.destination.name]
                if remap is None:
                    # don't add this connection back in
                    continue
                elif not isinstance(remap, basestring):
                    ops.extend(remap(old_conn, new_module))
                    continue
                else:
                    destination_name = remap
                    
            old_src_module = pipeline.modules[old_conn.source.moduleId]
            if use_registry:
                destination_port = destination_name
            else:
                destination_port = Port(name=destination_name,
                                        type='destination',
                                        signature=create_port_spec_string(
                                            [(basic_pkg, 'Variant', '')]))

            new_conn = create_new_connection(controller,
                                             old_src_module,
                                             old_conn.source,
                                             new_module,
                                             destination_port)
            ops.append(('add', new_conn))
        
        return [vistrails.core.db.action.create_action(ops)]

    @staticmethod
    def replace_group(controller, pipeline, module_id, new_subpipeline):
        basic_pkg = get_vistrails_basic_pkg_id()
        old_group = pipeline.modules[module_id]
        new_group = controller.create_module(basic_pkg, 'Group', '', 
                                             old_group.location.x, 
                                             old_group.location.y)
        new_group.pipeline = new_subpipeline
        return UpgradeWorkflowHandler.replace_generic(controller, pipeline, 
                                                      old_group, new_group)

    @staticmethod
    def replace_module(controller, pipeline, module_id, new_descriptor,
                       function_remap=None, src_port_remap=None,
                       dst_port_remap=None, annotation_remap=None,
                       control_param_remap=None, use_registry=True):
        old_module = pipeline.modules[module_id]
        internal_version = -1
        # try to determine whether new module is an abstraction
        if (hasattr(new_descriptor, 'module') and
            hasattr(new_descriptor.module, "vistrail") and 
            hasattr(new_descriptor.module, "internal_version")):
            internal_version = new_descriptor.version
        new_module = \
            controller.create_module_from_descriptor(new_descriptor,
                                                     old_module.location.x,
                                                     old_module.location.y,
                                                     internal_version,
                                                     not use_registry)

        return UpgradeWorkflowHandler.replace_generic(controller, pipeline, 
                                                      old_module, new_module,
                                                      function_remap, 
                                                      src_port_remap, 
                                                      dst_port_remap,
                                                      annotation_remap,
                                                      control_param_remap,
                                                      use_registry)

    @staticmethod
    def remap_module(controller, module_id, pipeline, pkg_remap):

        """remap_module offers a method to shortcut the
        specification of upgrades.  It is useful when just changing
        the names of ports or modules, but can also be used to add
        intermediate modules or change the format of parameters.  It
        is usually called from handle_module_upgrade_request, and the
        first three arguments are passed from the arguments to that
        method.

        pkg_remap specifies all of the changes and is of the format
        {<old_module_name>: [(<start_version>, <end_version>, 
                             <new_module_klass> | <new_module_id> | None, 
                             <remap_dictionary>)]}
        where new_module_klass is the class and new_module_id
        is a string of the format 
            <package_name>:[<namespace> | ]<module_name>
        passing None keeps the original name,
        and remap_dictionary is {<remap_type>:
        <name_changes>} and <name_changes> is a map from <old_name> to
        <new_name> or <remap_function>
        The remap functions are passed the old object and the new
        module and should return a list of operations with elements of
        the form ('add', <obj>).

        For example:

        def outputName_remap(old_conn, new_module):
            ops = []
            ...
            return ops
        pkg_remap = {'FileSink': [(None, '1.5.1', FileSink,
                                     {'dst_port_remap':
                                          {'overrideFile': 'overwrite',
                                           'outputName': outputName_remap},
                                      'function_remap':
                                          {'overrideFile': 'overwrite',
                                           'outputName': 'outputPath'}}),
                        }
        """

        reg = get_module_registry()

        old_module = pipeline.modules[module_id]
        old_version = old_module.version
        old_desc_str = create_descriptor_string(old_module.package,
                                                old_module.name,
                                                old_module.namespace,
                                                False)
        # print 'running module_upgrade_request', old_module.name
        if not isinstance(pkg_remap, UpgradePackageRemap):
            pkg_remap = UpgradePackageRemap.from_dict(pkg_remap)
        

        action_list = []

        old_module_t = \
            (old_module.package, old_module.name, old_module.namespace)
        module_remap = pkg_remap.get_module_upgrade(old_desc_str, old_version)
        tmp_pipeline = copy.copy(pipeline)
        while module_remap is not None:
            new_module_type = module_remap.new_module
            if new_module_type is None:
                new_module_t = old_module_t
            elif isinstance(new_module_type, basestring):
                new_module_t = parse_descriptor_string(new_module_type,
                                                       old_module_t[0])
            else:
                new_module_desc = reg.get_descriptor(new_module_type)
                new_module_t = new_module_desc.spec_tuple

            new_pkg_version = module_remap.output_version
            if (new_pkg_version is None or
                  reg.get_package_by_name(new_module_t[0]).version == new_pkg_version):
                # upgrading to the current version
                try:
                    new_module_desc = reg.get_descriptor_by_name(*new_module_t)
                except MissingModule, e:
                    # if the replacement is an abstraction,
                    # and it has been upgraded, we use that
                    if reg.has_abs_upgrade(*new_module_t):
                        new_module_desc = reg.get_abs_upgrade(*new_module_t)
                    else:
                        raise e
                use_registry = True
                next_module_remap = None
            else:
                new_module_desc = ModuleDescriptor(package=new_module_t[0],
                                                   name=new_module_t[1],
                                                   namespace=new_module_t[2],
                                                   package_version=new_pkg_version)
                use_registry = False

                # need to try more upgrades since this one isn't current
                old_desc_str = create_descriptor_string(new_module_t[0],
                                                        new_module_t[1],
                                                        new_module_t[2],
                                                        False)
                old_version = new_pkg_version
                next_module_remap = pkg_remap.get_module_upgrade(old_desc_str,
                                                                 old_version)
                old_module_t = new_module_t
            replace_module = UpgradeWorkflowHandler.replace_module
            actions = replace_module(controller, 
                                     tmp_pipeline,
                                     module_id, 
                                     new_module_desc,
                                     module_remap.function_remap,
                                     module_remap.src_port_remap,
                                     module_remap.dst_port_remap,
                                     module_remap.annotation_remap,
                                     module_remap.control_param_remap,
                                     use_registry)

            for a in actions:
                for op in a.operations:
                    # Update the id of the module being updated
                    if op.vtType == 'add' and op.what == 'module':
                        module_id = op.objectId
                tmp_pipeline.perform_action(a)

            action_list.extend(actions)
            module_remap = next_module_remap
        if len(action_list) > 0:
            return action_list

        # otherwise, just try to automatic upgrade
        # attempt_automatic_upgrade
        return UpgradeWorkflowHandler.attempt_automatic_upgrade(controller, 
                                                                pipeline,
                                                                module_id)
    
import unittest

class TestUpgradePackageRemap(unittest.TestCase):
    def test_from_dict(self):
        def outputName_remap(old_conn, new_module):
            ops = []
            return ops
        pkg_remap_d = {'FileSink': [(None, '1.5.1', None,
                                     {'dst_port_remap':
                                      {'overrideFile': 'overwrite',
                                       'outputName': outputName_remap},
                                      'function_remap':
                                      {'overrideFile': 'overwrite',
                                       'outputName': 'outputPath'}})]}
        pkg_remap = UpgradePackageRemap.from_dict(pkg_remap_d)

    def create_workflow(self, c):
        upgrade_test_pkg = 'org.vistrails.vistrails.tests.upgrade'

        d1 = ModuleDescriptor(package=upgrade_test_pkg,
                              name='TestUpgradeA',
                              namespace='',
                              package_version='0.8')
        m1 = c.create_module_from_descriptor(d1, use_desc_pkg_version=True)
        m1.is_valid = False
        c.add_module_action(m1)

        d2 = ModuleDescriptor(package=upgrade_test_pkg,
                              name='TestUpgradeB',
                              namespace='',
                              package_version = '0.8')
        m2 = c.create_module_from_descriptor(d2, use_desc_pkg_version=True)
        m2.is_valid = False
        c.add_module_action(m2)

        basic_pkg = get_vistrails_basic_pkg_id()
        psi = PortSpecItem(module="Float", package=basic_pkg,
                           namespace="", pos=0)
        function_port_spec = PortSpec(name="a", type="input", items=[psi])
        f = c.create_function(m1, function_port_spec, [12])
        c.add_function_action(m1, f)

        conn_out_psi = PortSpecItem(module="Integer", package=basic_pkg,
                                    namespace="", pos=0)
        conn_out_spec = PortSpec(name="z", type="output",
                                 items=[conn_out_psi])
        conn_in_psi = PortSpecItem(module="Integer", package=basic_pkg,
                                   namespace="", pos=0)
        conn_in_spec = PortSpec(name="b", type="input",
                                items=[conn_in_psi])
        conn = c.create_connection(m1, conn_out_spec, m2, conn_in_spec)
        c.add_connection_action(conn)
        return c.current_version

    def run_multi_upgrade_test(self, pkg_remap):
        from vistrails.core.application import get_vistrails_application

        app = get_vistrails_application()
        created_vistrail = False
        pm = get_package_manager()
        try:
            pm.late_enable_package('upgrades',
                                   {'upgrades':
                                    'vistrails.tests.resources.'})
            app.new_vistrail()
            created_vistrail = True
            c = app.get_controller()
            self.create_workflow(c)
        
            p = c.current_pipeline
            actions = UpgradeWorkflowHandler.remap_module(c, 0, p, pkg_remap)
        finally:
            if created_vistrail:
                app.close_vistrail()
            try:
                pm.late_disable_package('upgrades')
            except MissingPackage:
                pass

    def test_multi_upgrade_obj(self):
        module_remap_1 = UpgradeModuleRemap('0.8', '0.9', '0.9', None,
                                            module_name="TestUpgradeA")
        module_remap_1.add_remap('function_remap', 'a', 'aa')
        module_remap_1.add_remap('src_port_remap', 'z', 'zz')
        module_remap_2 = UpgradeModuleRemap('0.9', '1.0', '1.0', None,
                                            module_name="TestUpgradeA")
        module_remap_2.add_remap('function_remap', 'aa', 'aaa')
        module_remap_2.add_remap('src_port_remap', 'zz', 'zzz')
        pkg_remap = UpgradePackageRemap()
        pkg_remap.add_module_remap(module_remap_1)
        pkg_remap.add_module_remap(module_remap_2)
        self.run_multi_upgrade_test(pkg_remap)

    def test_multi_upgrade_dict(self):
        pkg_remap = {"TestUpgradeA": 
                     [UpgradeModuleRemap('0.8', '0.9', '0.9', None,
                                         function_remap={'a': 'aa'},
                                         src_port_remap={'z': 'zz'}),
                      UpgradeModuleRemap('0.9', '1.0', '1.0', None,
                                         function_remap={'aa': 'aaa'},
                                         src_port_remap={'zz': 'zzz'})]}
        self.run_multi_upgrade_test(pkg_remap)

    def test_multi_upgrade_legacy(self):
        # note that remap specifies the 0.8 -> 1.0 upgrade directly as
        # must be the case for legacy upgrades
        pkg_remap = {"TestUpgradeA": [('0.8', '1.0', None,
                                       {"function_remap": {'a': 'aaa'},
                                        "src_port_remap": {'z': 'zzz'}}),
                                      ('0.9', '1.0', None,
                                       {"function_remap": {'aa': 'aaa'},
                                        "src_port_remap": {'zz': 'zzz'}})]}
        self.run_multi_upgrade_test(pkg_remap)

    def test_multi_upgrade_rename(self):
        pkg_remap = {"TestUpgradeA": 
                     [UpgradeModuleRemap('0.8', '0.9', '0.9', "TestUpgradeB",
                                         dst_port_remap={'a': 'b'},
                                         src_port_remap={'z': 'zz'})],
                     "TestUpgradeB":
                     [UpgradeModuleRemap('0.9', '1.0', '1.0', None,
                                         src_port_remap={'zz': None})]}
        self.run_multi_upgrade_test(pkg_remap)

    def test_external_upgrade(self):
        from vistrails.core.application import get_vistrails_application

        app = get_vistrails_application()
        default_upgrades = app.temp_configuration.upgrades
        default_upgrade_delay = app.temp_configuration.upgradeDelay
        app.temp_configuration.upgrades = True
        app.temp_configuration.upgradeDelay = False

        created_vistrail = False
        pm = get_package_manager()
        try:
            pm.late_enable_package('upgrades',
                                   {'upgrades':
                                    'vistrails.tests.resources.'})
            app.new_vistrail()
            created_vistrail = True
            c = app.get_controller()
            current_version = self.create_workflow(c)
            for m in c.current_pipeline.modules.itervalues():
                self.assertEqual(m.version, '0.8')

            c.change_selected_version(current_version, from_root=True)
            
            self.assertEqual(len(c.current_pipeline.modules), 2)
            for m in c.current_pipeline.modules.itervalues():
                self.assertEqual(m.version, '1.0')
                if m.name == "TestUpgradeA":
                    self.assertEqual(m.functions[0].name, 'aaa')
            self.assertEqual(len(c.current_pipeline.connections), 1)
            conn = c.current_pipeline.connections.values()[0]
            self.assertEqual(conn.source.name, 'zzz')
            self.assertEqual(conn.destination.name, 'b')
                
        finally:
            if created_vistrail:
                app.close_vistrail()
            try:
                pm.late_disable_package('upgrades')
            except MissingPackage:
                pass
            app.temp_configuration.upgrades = default_upgrades
            app.temp_configuration.upgradeDelay = default_upgrade_delay

if __name__ == '__main__':
    import vistrails.core.application

    vistrails.core.application.init()
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2013-2014, NYU-Poly.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

try:
    import simplejson as json
except ImportError:
    import json

from vistrails.core.modules.vistrails_module import ModuleError

from ..common import TableObject, Table, InternalModuleError


class JSONTable(Table):
    _input_ports = [('file', '(org.vistrails.vistrails.basic:File)')]
    _output_ports = [
            ('column_count', '(org.vistrails.vistrails.basic:Integer)'),
            ('column_names', '(org.vistrails.vistrails.basic:List)'),
            ('value', Table)]

    def compute(self):
        json_file = self.get_input('file').name
        with open(json_file, 'rb') as fp:
            obj = json.load(fp)
        try:
            table = self.make_table(obj)
        except InternalModuleError, e:
            e.raise_module_error(self)
        self.set_output('column_count', table.columns)
        if table.names is not None:
            self.set_output('column_names', table.names)
        self.set_output('value', table)

    @staticmethod
    def add_list(columns, key, value):
        if not isinstance(value, list):
            raise InternalModuleError("Entry for key %r is not a list" % key)
        if key is not None:
            value = [key] + value
        for i, v in enumerate(value):
            columns[i].append(v)

    @staticmethod
    def add_dict(columns, keys, key_set, key, value):
        if not isinstance(value, dict):
            raise InternalModuleError(
                    "Entry for key %r is not an object" % key)
        value_keys = set(value.keys())
        m = key_set - value_keys
        if m:
            raise InternalModuleError(
                    "Entry for key %r is missing field %r" % (
                    key,
                    next(iter(m))))
        m = value_keys - key_set
        if m:
            raise InternalModuleError(
                    "Entry for key %r has unexpected field %r" % (
                    key,
                    next(iter(m))))
        if key is None:
            for i, k in enumerate(keys):
                columns[i].append(value[k])
        else:
            columns[0].append(key)
            for i, k in enumerate(keys):
                columns[i + 1].append(value[k])


class JSONObject(JSONTable):
    """Loads a JSON file and build a table from an object.

    In JSON, an object is written with {}. It is essentially an associative
    array. A column will contain the keys in this array.

    Example:
        {
            "John": {"lastname": "Smith", "age": 25, "city": "New York"},
            "Ashley": {"lastname": "Crofts", "age": 21, "city": "Fort Worth"},
            "Michael": {"lastname": "Buck", "age": 78, "city": "Goodman"}
        }

        key    | lastname | age |    city
        -------+----------+-----+-----------
        John   |  Smith   |  25 | New York
        Ashley |  Crofts  |  21 | Fort Worth
        Michal |  Buck    |  78 | Goodman

    Rows can also be lists (but they still all have to be in the same format).
    In this case, columns will not be named.

    To read a list of rows, use the JSONList module instead.
    """
    _input_ports = [('key_name', '(org.vistrails.vistrails.basic:String)',
                     {'optional': True, 'defaults': repr(["key"])})]

    def make_table(self, obj):
        if not isinstance(obj, dict):
            raise ModuleError(self, "JSON file is not an object")
        key_name = self.get_input('key_name', True)
        iterator = obj.iteritems()
        try:
            first_key, first_value = next(iterator)
        except StopIteration:
            raise ModuleError(self, "No entry in JSON object")
        count = 1
        if isinstance(first_value, list):
            keys = None
            columns = [[] for i in xrange(1 + len(first_value))]
            self.add_list(columns, first_key, first_value)
            for key, value in iterator:
                self.add_list(columns, key, value)
                count += 1
        elif isinstance(first_value, dict):
            keys = first_value.keys()
            key_set = set(keys)
            columns = [[] for i in xrange(1 + len(keys))]
            self.add_dict(columns, keys, key_set, first_key, first_value)
            for key, value in iterator:
                self.add_dict(columns, keys, key_set, key, value)
                count += 1
        else:
            raise ModuleError(self, "Values should be lists or objects")

        if keys is not None:
            names = [key_name] + keys
        else:
            names = None
        return TableObject(columns, count, names)


class JSONList(JSONTable):
    """Loads a JSON file and build a table from a list.

    In JSON, a list is written with [].

    Example:
        [[ 4, 14, 15,  1],
         [ 9,  7,  6, 12],
         [ 5, 11, 10,  8],
         [16,  2,  3, 13]]

        gives a 4x4 unnamed table.
    """
    def make_table(self, obj):
        if not isinstance(obj, list):
            raise ModuleError(self, "JSON file is not a list")
        iterator = iter(obj)
        try:
            first = next(iterator)
        except StopIteration:
            raise ModuleError(self, "No element in JSON list")
        count = 1
        if isinstance(first, list):
            keys = None
            columns = [[] for i in xrange(len(first))]
            self.add_list(columns, None, first)
            for value in iterator:
                self.add_list(columns, None, value)
                count += 1
        elif isinstance(first, dict):
            keys = first.keys()
            key_set = set(keys)
            columns = [[] for i in xrange(len(keys))]
            self.add_dict(columns, keys, key_set, None, first)
            for value in iterator:
                self.add_dict(columns, keys, key_set, None, value)
                count += 1
        else:
            raise ModuleError(self, "Values should be lists or objects")

        return TableObject(columns, count, keys)


_modules = [(JSONTable, {'abstract': True}), JSONObject, JSONList]


###############################################################################

import unittest
from vistrails.tests.utils import execute, intercept_results
from ..identifiers import identifier


class TestJSON(unittest.TestCase):
    def test_object(self):
        """Reads an object with object or list rows.
        """
        json_files = [
            ("""
            {
                "John": {"lastname": "Smith", "age": 25, "city": "New York"},
                "Lara": {"lastname": "Croft", "age": 21, "city": "Nashville"},
                "Michael": {"lastname": "Buck", "age": 78, "city": "Goodman"}
            }
            """, True),
            ("""
            {
                "John": ["Smith", 25, "New York"],
                "Lara": ["Croft", 21, "Nashville"],
                "Michael": ["Buck", 78, "Goodman"]
            }
            """, False),
            ]

        for json_file, has_names in json_files:
            with intercept_results(JSONObject, 'value', 'column_count',
                                   'column_names') as results:
                self.assertFalse(execute([
                        ('WriteFile', 'org.vistrails.vistrails.basic', [
                            ('in_value', [('String', json_file)]),
                        ]),
                        ('read|JSONObject', identifier, []),
                    ],
                    [
                        (0, 'out_value', 1, 'file'),
                    ]))
            self.assertTrue(all((len(r) == 1) for r in results[:2]))
            (table,), (count,), names = results
            self.assertEqual(count, 4)

            import numpy
            if has_names:
                self.assertEqual(names, [table.names])
                self.assertEqual(table.names[0], 'key')
                self.assertEqual(set(table.names[1:]),
                                 set(['lastname', 'age', 'city']))
                f_city = table.names.index('city')
                f_age = table.names.index('age')
            else:
                self.assertEqual(names, [])
                self.assertIsNone(table.names)
                f_city = 3
                f_age = 2
            self.assertEqual(set(table.get_column(f_city)),
                             set(["New York", "Nashville", "Goodman"]))
            l = table.get_column(f_age, True)
            self.assertIsInstance(l, numpy.ndarray)
            self.assertEqual(set(l), set([21, 25, 78]))

    def test_list(self):
        """Reads a list of object or list rows.
        """
        json_files = [
            """
            [
                {"firstname": "John", "lastname": "Smith", "age": 25},
                {"firstname": "Lara", "lastname": "Croft", "age": 21},
                {"firstname": "Michael", "lastname": "Buck", "age": 78}
            ]
            """,
            """
            [[2, 7, 6],
             [9, 5, 1],
             [4, 3, 8]]
            """,
            ]

        for nb, json_file in enumerate(json_files):
            with intercept_results(JSONList, 'value', 'column_count',
                                   'column_names') as results:
                self.assertFalse(execute([
                        ('WriteFile', 'org.vistrails.vistrails.basic', [
                            ('in_value', [('String', json_file)]),
                        ]),
                        ('read|JSONList', identifier, []),
                    ],
                    [
                        (0, 'out_value', 1, 'file'),
                    ]))
            self.assertTrue(all((len(r) == 1) for r in results[:2]))
            (table,), (count,), names = results
            self.assertEqual(count, 3)

            import numpy
            if nb == 0:
                self.assertEqual(names, [table.names])
                self.assertEqual(set(table.names),
                                 set(['firstname', 'lastname', 'age']))
                self.assertEqual(set(table.get_column_by_name('firstname')),
                                 set(["John", "Lara", "Michael"]))
                l = table.get_column_by_name('age', True)
                self.assertIsInstance(l, numpy.ndarray)
                self.assertEqual(set(l), set([21, 25, 78]))
            else:
                self.assertEqual(names, [])
                self.assertIsNone(table.names)
                self.assertEqual([table.get_column(col) for col in xrange(3)],
                                 [[2, 9, 4],
                                  [7, 5, 3],
                                  [6, 1, 8]])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from PyQt4 import QtCore, QtGui

from vistrails.core.system import get_vistrails_basic_pkg_id
from vistrails.gui.modules.module_configure import StandardModuleConfigurationWidget


class StringFormatConfigurationWidget(StandardModuleConfigurationWidget):
    """
    Configuration widget creating the ports corresponding to the format.

    """
    def __init__(self, module, controller, parent=None):
        """ StringFormatConfigurationWidget(
                module: Module,
                controller: VistrailController,
                parent: QWidget)
                -> TupleConfigurationWidget

        Let StandardModuleConfigurationWidget constructor store the
        controller/module object from the builder and set up the
        configuration widget.
        After StandardModuleConfigurationWidget constructor, all of
        these will be available:
        self.module : the Module object int the pipeline
        self.controller: the current vistrail controller

        """
        StandardModuleConfigurationWidget.__init__(self, module,
                                                   controller, parent)

        # Give it a nice window title
        self.setWindowTitle("StringFormat Configuration")

        # Add an empty vertical layout
        centralLayout = QtGui.QVBoxLayout()
        centralLayout.setMargin(0)
        centralLayout.setSpacing(0)
        self.setLayout(centralLayout)

        # Add the configuration button
        self.button = QtGui.QPushButton("Sync ports")
        self.connect(self.button, QtCore.SIGNAL('clicked()'),
                     self.saveTriggered)
        centralLayout.addWidget(self.button)

    def activate(self):
        self.button.focusWidget(QtCore.Qt.ActiveWindowFocusReason)

    def saveTriggered(self, checked = False):
        """ saveTriggered(checked: bool) -> None
        Update vistrail controller and module when the user click Ok

        """
        if self.updateVistrail():
            self.emit(QtCore.SIGNAL('stateChanged'))
            self.emit(QtCore.SIGNAL('doneConfigure'), self.module.id)

    def get_format(self):
        for i in xrange(self.module.getNumFunctions()):
            func = self.module.functions[i]
            if func.name == 'format':
                return func.params[0].strValue
        else:
            return ''

    def updateVistrail(self):
        """ updateVistrail() -> None
        Update Vistrail to contain changes in the port table

        """
        from vistrails.core.modules.basic_modules import StringFormat
        args, kwargs = StringFormat.list_placeholders(self.get_format())
        wanted_ports = set('_%d' % n for n in xrange(args)) | kwargs

        current_ports = set(port_spec.name
                            for port_spec in self.module.input_port_specs)

        sigstring = '(org.vistrails.vistrails.basic:Variant)'
        add_ports = [('input', n, sigstring, -1)
                     for n in (wanted_ports - current_ports)]
        delete_ports = [('input', n)
                        for n in (current_ports - wanted_ports)]

        self.controller.update_ports(self.module.id, delete_ports, add_ports)

        return True



#!/usr/bin/env python
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
# Basic information

import sys
import platform

print "Python:"
print "  Basic version: %s.%s.%s" % (sys.version_info[0],
                                     sys.version_info[1],
                                     sys.version_info[2], )
print "  Full version: " + sys.version.replace('\n', ' ')
print

def c(s):
    return s or "<COULD NOT DETERMINE>"

print "System:"
print "  Type: " + c(platform.system())
print "  Architecture: " + c(platform.architecture()[0])
print "  Machine: " + c(platform.machine())
print "  Platform: " + c(platform.platform())
print "  Processor: " + c(platform.processor())
print

##############################################################################

print "Libraries:"

try:
    import sip
    print "  sip installed."
    print "    version: " + sip.SIP_VERSION_STR
except ImportError:
    print "  sip NOT installed."
print

try:
    import PyQt4.Qt
    print "  PyQt installed."
    print "    Qt version: " + PyQt4.Qt.QT_VERSION_STR
    print "    PyQt version: " + PyQt4.Qt.PYQT_VERSION_STR
except ImportError:
    print "  PyQt NOT installed."
print

try:
    import vtk
    print "  VTK installed."
    print "    VTK short version: " + vtk.vtkVersion().GetVTKVersion()
    print "    VTK full version: " + vtk.vtkVersion().GetVTKSourceVersion()
except ImportError:
    print "  VTK NOT installed."



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This package defines a set of methods to deal with web services.
It requires ZSI library to be installed (please use the trunk version as it has
important fixes. Click on configure to add wsdl urls to the package 
(use a ; to separate the urls).

ChangeLog
2013-04-26  (by VisTrails Team)
    * Updated package to version 0.9.4
    * Updated to new org.vistrails identifiers
2010-02-25  (by VisTrails Team)
    * Updated package to version 0.9.3
    * Fixed bug with new package reloading
    * If you see the error ('Error loading configuration file: ',
                             '~/.vistrails/webServices/modules.conf')
      Just remove your ~/.vistrails/webServices and enable the package again
    * Adding http package to package dependencies
2010-01-27  (by VisTrails Team)
    * Updated package to version 0.9.2 
    * Supporting hierarchy of types (not fully tested yet)
    * Changing the default port names of modules to 'self', keeping the old 
      names (as optional ports) for backwards compatibility 
2010-01-25  (by VisTrails Team)
    * Updated package to version 0.9.1 
    * Expanded map of simple types
"""
from __future__ import division

from vistrails.core.configuration import ConfigurationObject
import vistrails.core

identifier = 'org.vistrails.vistrails.webservices'
name = 'Web Services'
version = '0.9.4'
old_identifiers = ['edu.utah.sci.vistrails.webservices']
configuration = ConfigurationObject(wsdlList=(None, str),
                                    showWarning=True)

def package_dependencies():
    return ['org.vistrails.vistrails.http']
    
def package_requirements():
    from vistrails.core.requirements import require_python_module
    require_python_module('ZSI', {
            'pip': 'zsi',
            'linux-debian': 'python-zsi',
            'linux-ubuntu': 'python-zsi',
            'linux-fedora': 'python-ZSI'})



from __future__ import division

try:
    import numpy
except ImportError: # pragma: no cover
    numpy = None
import re

from vistrails.core.modules.vistrails_module import ModuleError

from .common import TableObject, Table, choose_column, choose_columns

# FIXME use pandas?


def utf8(obj):
    if isinstance(obj, bytes):
        return obj
    elif isinstance(obj, unicode):
        return obj.encode('utf-8')
    else:
        return bytes(obj)


class JoinedTables(TableObject):
    def __init__(self, left_t, right_t, left_key_col, right_key_col,
                 case_sensitive=False, always_prefix=False):
        self.left_t = left_t
        self.right_t = right_t
        self.left_key_col = left_key_col
        self.right_key_col = right_key_col
        self.case_sensitive = case_sensitive
        self.always_prefix = always_prefix

        self.build_column_names()
        self.compute_row_map()
        self.column_cache = {}
        self.rows = len(self.row_map)

    def build_column_names(self):
        left_name = self.left_t.name
        if left_name is None:
            left_name = "left"
        right_name = self.right_t.name
        if right_name is None:
            right_name = "right"

        def get_col_names(table, other, prefix):
            names = []
            for i in xrange(table.columns):
                should_prefix = self.always_prefix
                if table.names is not None:
                    n = table.names[i]
                else:
                    n = "col %d" % i
                    should_prefix = True
                if not should_prefix and other.names is not None:
                    if n in other.names:
                        should_prefix = True
                names.append("%s%s" % (prefix + "." if should_prefix else "",
                                       n))
            return names

        self.names = (get_col_names(self.left_t, self.right_t, left_name) +
                      get_col_names(self.right_t, self.left_t, right_name))
        self.columns = len(self.names)

    def get_column(self, index, numeric=False):
        if (index, numeric) in self.column_cache:
            return self.column_cache[(index, numeric)]

        result = []
        if index < self.left_t.columns:
            column = self.left_t.get_column(index, numeric)
            for i in xrange(self.left_t.rows):
                if i in self.row_map:
                    result.append(column[i])
        else:
            column = self.right_t.get_column(index - self.left_t.columns,
                                             numeric)
            for i in xrange(self.left_t.rows):
                if i in self.row_map:
                    j = self.row_map[i]
                    result.append(column[j])

        if numeric and numpy is not None:
            result = numpy.array(result, dtype=numpy.float32)
        self.column_cache[(index, numeric)] = result
        return result

    def compute_row_map(self):
        def build_key_dict(table, key_col):
            column = table.get_column(key_col)
            if self.case_sensitive:
                key_dict = dict((utf8(val).strip(), i)
                                for i, val in enumerate(column))
            else:
                key_dict = dict((utf8(val).strip().upper(), i)
                                for i, val in enumerate(column))
            return key_dict

        right_keys = build_key_dict(self.right_t, self.right_key_col)

        self.row_map = {}
        for left_row_idx, key in enumerate(
                self.left_t.get_column(self.left_key_col)):
            key = utf8(key).strip()
            if not self.case_sensitive:
                key = key.upper()
            if key in right_keys:
                self.row_map[left_row_idx] = right_keys[key]


class JoinTables(Table):
    """Joins data from two tables using equality of a pair of columns.

    This creates a table by combining the fields from the two tables. It will
    match the values in the two selected columns (one from each table). If a
    row from one of the table has a value for the selected field that doesn't
    exist in the other table, that row will not appear in the result
    (INNER JOIN semantics).
    """
    _input_ports = [('left_table', 'Table'),
                    ('right_table', 'Table'),
                    ('left_column_idx', 'basic:Integer'),
                    ('left_column_name', 'basic:String'),
                    ('right_column_idx', 'basic:Integer'),
                    ('right_column_name', 'basic:String'),
                    ('case_sensitive', 'basic:Boolean',
                     {"optional": True, "defaults": str(["False"])}),
                    ('always_prefix', 'basic:Boolean',
                     {"optional": True, "defaults": str(["False"])})]
    _output_ports = [('value', Table)]

    def compute(self):
        left_t = self.get_input('left_table')
        right_t = self.get_input('right_table')
        case_sensitive = self.get_input('case_sensitive')
        always_prefix = self.get_input('always_prefix')

        def get_column_idx(table, prefix):
            col_name_port = "%s_column_name" % prefix
            col_idx_port = '%s_column_idx' % prefix
            try:
                col_idx = choose_column(
                        table.columns,
                        column_names=table.names,
                        name=self.force_get_input(col_name_port, None),
                        index=self.force_get_input(col_idx_port, None))
            except ValueError, e:
                raise ModuleError(self, e.message)

            return col_idx

        left_key_col = get_column_idx(left_t, "left")
        right_key_col = get_column_idx(right_t, "right")

        table = JoinedTables(left_t, right_t, left_key_col, right_key_col,
                             case_sensitive, always_prefix)
        self.set_output('value', table)


class ProjectedTable(TableObject):
    def __init__(self, table, col_idxs, col_names):
        self.table = table
        self.col_map = dict(enumerate(col_idxs))
        self.columns = len(self.col_map)
        self.names = col_names

    def get_column(self, index, numeric=False):
        mapped_idx = self.col_map[index]
        return self.table.get_column(mapped_idx, numeric)

    @property
    def rows(self):
        return self.table.rows


class ProjectTable(Table):
    """Build a table from the columns of another table.

    This allows you to restrict, reorder or rename the columns of a table. You
    can also duplicate columns by mentioning them several times.
    """
    _input_ports = [("table", "Table"),
                    ("column_names", "basic:List"),
                    ("column_indexes", "basic:List"),
                    ("new_column_names", "basic:List",
                     {"optional": True})]
    _output_ports = [("value", Table)]

    def compute(self):
        table = self.get_input("table")
        try:
            indexes = choose_columns(
                    table.columns,
                    column_names=table.names,
                    names=self.force_get_input('column_names', None),
                    indexes=self.force_get_input('column_indexes', None))
        except ValueError, e:
            raise ModuleError(self, e.message)
        if self.has_input('new_column_names'):
            column_names = self.get_input('new_column_names')
            if len(column_names) != len(indexes):
                raise ModuleError(self,
                                  "new_column_names was specified but doesn't "
                                  "have the right number of names")
        else:
            column_names = []
            names = {}
            for i in indexes:
                name = table.names[i]
                if name in names:
                    nb = names[name]
                    names[name] += 1
                    name = '%s_%d' % (name, nb)
                else:
                    names[name] = 1
                column_names.append(name)

        projected_table = ProjectedTable(table, indexes, column_names)
        self.set_output("value", projected_table)


class SelectFromTable(Table):
    """Builds a table from the rows of another table.

    This allows you to filter the records in a table according to a condition
    on a specific field.
    """
    _input_ports = [('table', 'Table'),
                    ('str_expr', 'basic:String,basic:String,basic:String',
                     {'entry_types': "['default','enum','default']",
                      'values': "[[], ['==', '!=', '=~'], []]"}),
                    ('float_expr', 'basic:String,basic:String,basic:Float',
                     {'entry_types': "['default','enum','default']",
                      'values': "[[], ['==', '!=', '<', '>', '<=', '>='], []]"})]
    _output_ports = [('value', Table)]

    @staticmethod
    def make_condition(comparand, comparer):
        if isinstance(comparand, float):
            with_cast = lambda f: lambda v: f(float(v))
        else:
            with_cast = lambda f: f
        if comparer == '==':
            return with_cast(lambda v: v == comparand)
        elif comparer == '!=':
            return with_cast(lambda v: v != comparand)
        elif comparer == '<':
            return with_cast(lambda v: v < comparand)
        elif comparer == '>':
            return with_cast(lambda v: v > comparand)
        elif comparer == '<=':
            return with_cast(lambda v: v <= comparand)
        elif comparer == '>=':
            return with_cast(lambda v: v >= comparand)
        elif comparer == '=~':
            regex = re.compile(comparand)
            return regex.search
        else:
            raise ValueError("Invalid comparison operator %r" % comparer)

    def compute(self):
        table = self.get_input('table')

        if self.has_input('str_expr'):
            (col, comparer, comparand) = self.get_input('str_expr')
        elif self.has_input('float_expr'):
            (col, comparer, comparand) = self.get_input('float_expr')
        else:
            raise ModuleError(self, "Must have some expression")

        try:
            idx = int(col)
        except ValueError:
            try:
                idx = table.names.index(col)
            except ValueError:
                raise ModuleError(self, "No column %r" % col)
        else:
            if idx < 0 or idx >= table.columns:
                raise ModuleError(self,
                                  "No column %d, table only has %d columns" % (
                                  idx, table.columns))

        condition = self.make_condition(comparand, comparer)
        numeric = isinstance(comparand, float)
        column = table.get_column(idx, numeric)
        matched_rows = [i
                        for i, col_val in enumerate(column)
                        if condition(col_val)]
        columns = []
        for col in xrange(table.columns):
            column = table.get_column(col)
            columns.append([column[row] for row in matched_rows])
        selected_table = TableObject(columns, len(matched_rows), table.names)
        self.set_output('value', selected_table)


class AggregatedTable(TableObject):
    def __init__(self, table, op, col, group_col):
        self.table = table
        self.op = op
        self.col = col
        self.group_col = group_col

        self.build_map()

    def build_map(self):
        agg_map = {}
        for i, val in enumerate(self.table.get_column(self.group_col)):
            if val in agg_map:
                agg_map[val].append(i)
            else:
                agg_map[val] = [i]
        self.agg_rows = [(min(rows), rows) for rows in agg_map.itervalues()]
        self.agg_rows.sort()
        self.rows = len(self.agg_rows)
        self.columns = 2
        if self.table.names is not None:
            self.names = [self.table.names[self.group_col],
                          self.table.names[self.col]]

    def get_column(self, index, numeric=False):
        def average(value_iter):
            # value_iter can only be used once
            sum = 0
            count = 0
            for count, v in enumerate(value_iter):
                sum += v
            return sum / (count+1)
        op_map = {'sum': sum,
                  'average': average,
                  'min': min,
                  'max': max}
        if index == 0:
            col = self.table.get_column(self.group_col, numeric)
            return [col[x[0]] for x in self.agg_rows]
        else:
            if self.op == 'count':
                return [len(x[1]) for x in self.agg_rows]
            elif self.op in op_map:
                col = self.table.get_column(self.col, True)
                return [op_map[self.op](col[idx] for idx in x[1])
                        for x in self.agg_rows]
            else:
                raise ValueError('Unknown operation: "%s"' % self.op)


class AggregateColumn(Table):
    _input_ports = [('table', 'Table'),
                    ('op', 'basic:String',
                     {'entry_types': "['enum']",
                      'values': "[['sum', 'count', 'average', 'min', 'max']]"}),
                    ('column_name', 'basic:String'),
                    ('column_index', 'basic:Integer'),
                    ('group_by_name', 'basic:String'),
                    ('group_by_index', 'basic:Integer')]
    _output_ports = [('value', 'Table')]

    def compute(self):
        table = self.get_input('table')
        op = self.get_input('op')
        column_name = self.force_get_input('column_name', None)
        column_index = self.force_get_input('column_index', None)
        col_idx = choose_column(table.columns,
                                column_names=table.names,
                                name=column_name,
                                index=column_index)
        group_by_name = self.force_get_input('group_by_name', None)
        group_by_index = self.force_get_input('group_by_index', None)
        gb_idx = choose_column(table.columns,
                               column_names=table.names,
                               name=group_by_name,
                               index=group_by_index)

        res_table = AggregatedTable(table, op, col_idx, gb_idx)
        self.set_output('value', res_table)

_modules = [JoinTables, ProjectTable, SelectFromTable, AggregateColumn]


###############################################################################

import unittest
from vistrails.tests.utils import execute, intercept_result
from .identifiers import identifier


class TestJoin(unittest.TestCase):
    def test_join(self):
        """Test joining tables that have column names.
        """
        with intercept_result(JoinTables, 'value') as results:
            self.assertFalse(execute([
                    ('BuildTable', identifier, [
                        ('id', [('List', repr([1, '2', 4, 5]))]),
                        ('A_name', [('List',
                            repr(['one', 2, 'four', 'five'])),
                        ]),
                    ]),
                    ('BuildTable', identifier, [
                        ('B_age', [('List',
                            repr([14, 50, '12', 22])),
                        ]),
                        ('id', [('List', repr(['1', 2, 3, 5]))]),
                    ]),
                    ('JoinTables', identifier, [
                        ('left_column_idx', [('Integer', '0')]),
                        ('right_column_name', [('String', 'id')]),
                        ('right_column_idx', [('Integer', '1')]),
                    ]),
                ],
                [
                    (0, 'value', 2, 'left_table'),
                    (1, 'value', 2, 'right_table'),
                ],
                add_port_specs=[
                    (0, 'input', 'id',
                     'org.vistrails.vistrails.basic:List'),
                    (0, 'input', 'A_name',
                     'org.vistrails.vistrails.basic:List'),
                    (1, 'input', 'B_age',
                     'org.vistrails.vistrails.basic:List'),
                    (1, 'input', 'id',
                     'org.vistrails.vistrails.basic:List'),
                ]))
        self.assertEqual(len(results), 1)
        table, = results

        self.assertEqual(table.names, ['left.id', 'A_name',
                                       'B_age', 'right.id'])

        self.assertEqual(table.get_column(0, False), [1, '2', 5])
        l = table.get_column(0, True)
        self.assertIsInstance(l, numpy.ndarray)
        self.assertEqual(list(l), [1, 2, 5])
        self.assertEqual(table.get_column(3, False), ['1', 2, 5])
        l = table.get_column(3, True)
        self.assertIsInstance(l, numpy.ndarray)
        self.assertEqual(list(l), [1, 2, 5])

        self.assertEqual(table.get_column(1, False), ['one', 2, 'five'])
        self.assertEqual(list(table.get_column(2, True)), [14, 50, 22])

    def test_noname(self):
        """Tests joining tables that have no column names.
        """
        with intercept_result(JoinTables, 'value') as results:
            self.assertFalse(execute([
                    ('WriteFile', 'org.vistrails.vistrails.basic', [
                        ('in_value', [('String', '1;one\n2;2\n4;four\n'
                                                 '5;five')]),
                    ]),
                    ('read|CSVFile', identifier, [
                        ('delimiter', [('String', ';')]),
                        ('header_present', [('Boolean', 'False')]),
                        ('sniff_header', [('Boolean', 'False')]),
                    ]),
                    ('WriteFile', 'org.vistrails.vistrails.basic', [
                        ('in_value', [('String', '14;1\n50;2\n12;3\n22;5\n')]),
                    ]),
                    ('read|CSVFile', identifier, [
                        ('delimiter', [('String', ';')]),
                        ('header_present', [('Boolean', 'False')]),
                        ('sniff_header', [('Boolean', 'False')]),
                    ]),
                    ('JoinTables', identifier, [
                        ('left_column_idx', [('Integer', '0')]),
                        ('right_column_idx', [('Integer', '1')]),
                    ]),
                ],
                [
                    (0, 'out_value', 1, 'file'),
                    (2, 'out_value', 3, 'file'),
                    (1, 'value', 4, 'left_table'),
                    (3, 'value', 4, 'right_table'),
                ]))
        self.assertEqual(len(results), 1)
        table, = results

        self.assertEqual(table.names, ['left.col 0', 'left.col 1',
                                       'right.col 0', 'right.col 1'])
        self.assertEqual(table.get_column(0, False), ['1', '2', '5'])
        self.assertEqual(table.get_column(1, False), ['one', '2', 'five'])


class TestProjection(unittest.TestCase):
    def do_project(self, project_functions, error=None):
        with intercept_result(ProjectTable, 'value') as results:
            errors = execute([
                    ('BuildTable', identifier, [
                        ('letters', [('List', repr(['a', 'b', 'c', 'd']))]),
                        ('numbers', [('List', repr([1, 2, 3, '4']))]),
                        ('cardinals', [('List', repr(['one', 'two',
                                                      'three', 'four']))]),
                        ('ordinals', [('List', repr(['first', 'second',
                                                     'third', 'fourth']))])
                    ]),
                    ('ProjectTable', identifier, project_functions),
                ],
                [
                    (0, 'value', 1, 'table'),
                ],
                add_port_specs=[
                    (0, 'input', 'letters',
                     'org.vistrails.vistrails.basic:List'),
                    (0, 'input', 'numbers',
                     'org.vistrails.vistrails.basic:List'),
                    (0, 'input', 'cardinals',
                     'org.vistrails.vistrails.basic:List'),
                    (0, 'input', 'ordinals',
                     'org.vistrails.vistrails.basic:List'),
                ])
        if error is not None:
            self.assertEqual([1], errors.keys())
            self.assertIn(error, errors[1].message)
            return None
        else:
            self.assertFalse(errors)
            self.assertEqual(len(results), 1)
            return results[0]

    def test_column_numbers(self):
        """Projects using column numbers.
        """
        self.do_project([
                ('column_indexes', [('List', '[0, 4, 1, 0]')]),
            ],
            'table only has 4 columns')
        result = self.do_project([
                ('column_indexes', [('List', '[0, 3, 1, 0]')]),
            ])
        self.assertEqual(result.names, ['letters', 'ordinals',
                                        'numbers', 'letters_1'])
        self.assertEqual(result.get_column(0, False),
                         ['a', 'b', 'c', 'd'])
        self.assertEqual(list(result.get_column(2, True)),
                         [1, 2, 3, 4])
        self.assertEqual(result.get_column(3, False),
                         ['a', 'b', 'c', 'd'])

    def test_column_numbers_rename(self):
        """Projects and rename columns, using column numbers.
        """
        self.do_project([
                ('column_indexes', [('List', '[0, 3, 1, 0]')]),
                ('new_column_names', [('List', '["a", "b", "c"]')])
            ],
            "doesn't have the right number of names")
        result = self.do_project([
                ('column_indexes', [('List', '[0, 3, 1, 0]')]),
                ('new_column_names', [('List', '["a", "b", "c", "d"]')]),
            ])
        self.assertEqual(result.names, ['a', 'b', 'c', 'd'])

    def test_column_names(self):
        """Projects using column names.
        """
        self.do_project([
                ('column_names', [('List', repr(['letters', 'crackers']))]),
            ],
            "not found: 'crackers'")
        self.do_project([
                ('column_names', [('List', repr(['letters', 'ordinals']))]),
                ('column_indexes', [('List', '[0, 2]')]),
            ],
            "they don't agree")
        self.do_project([
                ('column_names', [('List', repr(['letters', 'ordinals']))]),
                ('column_indexes', [('List', '[0, 3]')]),
            ])
        result = self.do_project([
                ('column_names', [('List', repr(['letters', 'ordinals',
                                                 'letters']))]),
            ])
        self.assertEqual(result.names, ['letters', 'ordinals', 'letters_1'])


class TestSelect(unittest.TestCase):
    def do_select(self, select_functions, error=None):
        with intercept_result(SelectFromTable, 'value') as results:
            errors = execute([
                    ('WriteFile', 'org.vistrails.vistrails.basic', [
                        ('in_value', [('String', '22;a;T;abaab\n'
                                                 '43;b;F;aabab\n'
                                                 '-7;d;T;abbababb\n'
                                                 '500;e;F;aba abacc')]),
                    ]),
                    ('read|CSVFile', identifier, [
                        ('delimiter', [('String', ';')]),
                        ('header_present', [('Boolean', 'False')]),
                        ('sniff_header', [('Boolean', 'False')]),
                    ]),
                    ('SelectFromTable', identifier, select_functions),
                ],
                [
                    (0, 'out_value', 1, 'file'),
                    (1, 'value', 2, 'table'),
                ])
        if error is not None:
            self.assertEqual([2], errors.keys())
            self.assertIn(error, errors[2].message)
            return None
        else:
            self.assertFalse(errors)
            self.assertEqual(len(results), 1)
            return results[0]

    def test_numeric(self):
        """Selects using the 'less-than' condition.
        """
        self.do_select([
                ('float_expr', [('String', '6'),
                                ('String', '<='),
                                ('Float', '42.0')]),
            ],
            "table only has 4 columns")
        table = self.do_select([
                ('float_expr', [('String', '0'),
                                ('String', '<='),
                                ('Float', '42.0')]),
            ])
        l = table.get_column(0, True)
        self.assertIsInstance(l, numpy.ndarray)
        self.assertEqual(list(l), [22, -7])
        self.assertEqual(table.get_column(1, False), ['a', 'd'])

    def test_text(self):
        """Selects using the 'equal' condition.
        """
        table = self.do_select([
                ('str_expr', [('String', '2'),
                              ('String', '=='),
                              ('String', 'T')])
            ])
        self.assertEqual(table.get_column(0, False), ['22', '-7'])

    def test_regex(self):
        """Selects using the 'regex-match' condition.
        """
        table = self.do_select([
                ('str_expr', [('String', '3'),
                              ('String', '=~'),
                              ('String', r'([ab])\1')]),
            ])
        self.assertEqual(table.get_column(0, False), ['22', '43', '-7'])

class TestAggregate(unittest.TestCase):
    def do_aggregate(self, agg_functions):
        with intercept_result(AggregateColumn, 'value') as results:
            errors = execute([
                    ('WriteFile', 'org.vistrails.vistrails.basic', [
                        ('in_value', [('String', '22;a;T;100\n'
                                                 '43;b;F;3\n'
                                                 '-7;d;T;41\n'
                                                 '500;e;F;21\n'
                                                 '20;a;T;1\n'
                                                 '43;b;F;23\n'
                                                 '21;a;F;41\n')]),
                    ]),
                    ('read|CSVFile', identifier, [
                        ('delimiter', [('String', ';')]),
                        ('header_present', [('Boolean', 'False')]),
                        ('sniff_header', [('Boolean', 'False')]),
                    ]),
                    ('AggregateColumn', identifier, agg_functions),
                ],
                [
                    (0, 'out_value', 1, 'file'),
                    (1, 'value', 2, 'table'),
                ])
        self.assertFalse(errors)
        self.assertEqual(len(results), 1)
        return results[0]

    def test_aggregate_sum(self):
        table = self.do_aggregate([('op', [('String', 'sum')]),
                                   ('column_index', [('Integer', '0')]),
                                   ('group_by_index', [('Integer', '1')])])
        self.assertEqual(table.get_column(0, False), ['a', 'b', 'd', 'e'])
        self.assertEqual(table.get_column(1, True), [63, 86, -7, 500])

    def test_aggregate_avg(self):
        table = self.do_aggregate([('op', [('String', 'average')]),
                                   ('column_index', [('Integer', '3')]),
                                   ('group_by_index', [('Integer', '2')])])
        self.assertEqual(table.get_column(0, False), ['T', 'F'])
        second_col = list(table.get_column(1, True))
        self.assertAlmostEqual(second_col[0], 47.3333333333333)
        self.assertAlmostEqual(second_col[1], 22.0)

    def test_aggregate_min(self):
        table = self.do_aggregate([('op', [('String', 'min')]),
                                   ('column_index', [('Integer', '0')]),
                                   ('group_by_index', [('Integer', '2')])])
        self.assertEqual(table.get_column(0, False), ['T', 'F'])
        self.assertEqual(table.get_column(1, True), [-7, 21])



#!/usr/bin/env python

# This example shows how to color an isosurface with other
# data. Basically an isosurface is generated, and a data array is
# selected and used by the mapper to color the surface.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Read some data. The important thing here is to read a function as a
# data array as well as the scalar and vector.  (here function 153 is
# named "Velocity Magnitude").Later this data array will be used to
# color the isosurface.
pl3d = vtk.vtkPLOT3DReader()
pl3d.SetXYZFileName(VTK_DATA_ROOT + "/Data/combxyz.bin")
pl3d.SetQFileName(VTK_DATA_ROOT + "/Data/combq.bin")
pl3d.SetScalarFunctionNumber(100)
pl3d.SetVectorFunctionNumber(202)
pl3d.AddFunction(153)
pl3d.Update()
pl3d.DebugOn()
    
# The contour filter uses the labeled scalar (function number 100
# above to generate the contour surface; all other data is
# interpolated during the contouring process.
iso = vtk.vtkContourFilter()
iso.SetInputConnection(pl3d.GetOutputPort())
iso.SetValue(0, .24)

normals = vtk.vtkPolyDataNormals()
normals.SetInputConnection(iso.GetOutputPort())
normals.SetFeatureAngle(45)

# We indicate to the mapper to use the velcoity magnitude, which is a
# vtkDataArray that makes up part of the point attribute data.
isoMapper = vtk.vtkPolyDataMapper()
isoMapper.SetInputConnection(normals.GetOutputPort())
isoMapper.ScalarVisibilityOn()
isoMapper.SetScalarRange(0, 1500)
isoMapper.SetScalarModeToUsePointFieldData()
isoMapper.ColorByArrayComponent("VelocityMagnitude", 0)

isoActor = vtk.vtkLODActor()
isoActor.SetMapper(isoMapper)
isoActor.SetNumberOfCloudPoints(1000)

outline = vtk.vtkStructuredGridOutlineFilter()
outline.SetInputConnection(pl3d.GetOutputPort())
outlineMapper = vtk.vtkPolyDataMapper()
outlineMapper.SetInputConnection(outline.GetOutputPort())
outlineActor = vtk.vtkActor()
outlineActor.SetMapper(outlineMapper)

# Create the usual rendering stuff.
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(outlineActor)
ren.AddActor(isoActor)
ren.SetBackground(1, 1, 1)
renWin.SetSize(500, 500)
ren.SetBackground(0.1, 0.2, 0.4)

cam1 = ren.GetActiveCamera()
cam1.SetClippingRange(3.95297, 50)
cam1.SetFocalPoint(9.71821, 0.458166, 29.3999)
cam1.SetPosition(2.7439, -37.3196, 38.7167)
cam1.SetViewUp(-0.16123, 0.264271, 0.950876)

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This module contains class definitions for:

    * PEFunction
"""
from __future__ import division

from vistrails.db.domain import DBPEFunction, DBPEParameter
from vistrails.core.paramexplore.param import PEParam
import copy

import unittest
from vistrails.db.domain import IdScope

################################################################################

class PEFunction(DBPEFunction):
    """ Stores a function for a parameter exploration """

    ##########################################################################
    # Constructors and copy
    
    def __init__(self, *args, **kwargs):
        DBPEFunction.__init__(self, *args, **kwargs)
        if self.id is None:
            self.id = -1
        if self.module_id is None:
            self.module_id = -1
        if self.port_name is None:
            self.port_name = ""
        if self.is_alias is None:
            self.is_alias = 0
        self.set_defaults()
        
    def set_defaults(self, other=None):
        pass

    def __copy__(self):
        """ __copy__() -> ModuleFunction - Returns a clone of itself """
        return PEFunction.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBPEFunction.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = PEFunction
        cp.set_defaults(self)
        return cp

    @staticmethod
    def convert(_function):
        if _function.__class__ == PEFunction:
            return
        _function.__class__ = PEFunction
        for _parameter in _function.db_get_parameters():
            PEParam.convert(_parameter)
        _function.set_defaults()

    ##########################################################################
    # Properties

    id = DBPEFunction.db_id
    module_id = DBPEFunction.db_module_id
    port_name = DBPEFunction.db_port_name
    is_alias = DBPEFunction.db_is_alias

    def _get_params(self):
        self.db_parameters.sort(key=lambda x: x.db_pos)
        return self.db_parameters
    def _set_params(self, params):
        self.db_parameters = params

    parameters = property(_get_params, _set_params)
    params = property(_get_params, _set_params)

    def add_parameter(self, param):
        self.db_add_parameter(param)
    addParameter = add_parameter

    def add_parameters(self, params):
        for p in params:
            self.db_add_parameter(p)

    ##########################################################################
    # Operators
    
    def __str__(self):
        """ __str__() -> str - Returns a string representation of itself """
        return ("<PEFunction id=%s module_id=%s port_name='%s' is_alias=%s params=%s)@%X" %
                (self.id,
                 self.module_id,
                 self.port_name,
                 self.is_alias,
                 [str(p) for p in self.params],
                 id(self)))

    def __eq__(self, other):
        """ __eq__(other: ModuleFunction) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if type(self) != type(other):
            return False
        if self.port_name != other.port_name:
            return False
        if self.module_id != other.module_id:
            return False
        if self.is_alias != other.is_alias:
            return False
        if len(self.params) != len(other.params):
            return False
        for p,q in zip(self.params, other.params):
            if p != q:
                return False
        return True
            
    def __ne__(self, other):
        """ __ne__(other: ModuleFunction) -> boolean
        Returns True if self and other don't have the same attributes. 
        Used by !=  operator. 
        
        """
        return not self.__eq__(other)

################################################################################
# Testing

class TestModuleFunction(unittest.TestCase):

    def create_function(self, id_scope=IdScope()):
        param = PEParam(id=id_scope.getNewId(PEParam.vtType),
                            pos=2,
                            interpolator='Stepper',
                            dimension=4,
                            value='[1, 2]')
        function = PEFunction(id=id_scope.getNewId(PEFunction.vtType),
                                  module_id=7,
                                  port_name='value',
                                  is_alias=0,
                                  parameters=[param])
        return function

    def test_copy(self):        
        id_scope = IdScope()
        f1 = self.create_function(id_scope)
        f2 = copy.copy(f1)
        self.assertEquals(f1, f2)
        self.assertEquals(f1.id, f2.id)
        f3 = f1.do_copy(True, id_scope, {})
        self.assertEquals(f1, f3)
        self.assertNotEquals(f1.id, f3.id)

    def testComparisonOperators(self):
        f = self.create_function()
        g = self.create_function()
        self.assertEqual(f, g)
        g.module_id = 1 
        self.assertNotEqual(f, g)
        g.module_id = 7
        g.port_name = "val"
        self.assertNotEqual(f, g)
        g.port_name = "value"
        g.is_alias = 1
        self.assertNotEqual(f, g)

    def test_str(self):
        str(self.create_function())

if __name__ == '__main__':
    unittest.main()
    



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
# ################################################################################
# # QueryWF: Base class for queries over workflows

# class QueryWF(object):

#     def __init__(self, transformer, logQueries):
#         self.transformer = transformer
#         self. = logQueries
    

# ################################################################################

# class QueryWFModuleName(QueryWF):
    
#     def __init__(self, name):
#         self.name = name

#     def run(self, module):
#         return module.name == self.name

# ################################################################################

# class QueryWFParameter(QueryWF):
    
#     def __init__(self, param, name):
#         self.param = param
#         self.name = name

#     def run(self, module):
#         for f in module.functions:
#             if (f.name == name and 
#                 len(f.params) == 1 and
#                 f.params[0].name == self.name:
#                 return True
#         return False

# ################################################################################

# class QueryWFAnd(QueryWF):
            
#     def __init__(self, queries):
#         self.queries = queries

#     def run(self, module):
#         for query in self.queries:
#             if not query.run(module):
#                 return False
#         return True

# ################################################################################

# class QueryWFOr(QueryWF):
    
#     def __init__(self, queries):
#         self.queries = queries

#     def run(self, module):
#         for query in self.queries:
#             if query.run(module):
#                 return True
#         return False

# ################################################################################

# class QueryWFNot(QueryWF):

#     def __init__(self, query):
#         self.query = query

#     def run(self, module):
#         return not self.query.run(module)

# ################################################################################

# class QueryWFResultTransformer(object):
#     pass

# ################################################################################

# class QueryWFUpstream(QueryWFResultTransformer):

#     def run(self, pipeline, module_ids):
#         result = set()
#         inv_graph = pipeline.graph.inverse()
#         for i in module_ids:
#             result = result.union(set(inv_graph.bfs(i).keys() + [i]))
#         return result


# class QueryWFIdentity(QueryWFResultTransformer):

#     def run(self, pipeline, module_ids):
#         return module_ids


# class QueryWFUnion(QueryWFResultTransformer):

#     def __init__(self, queries):
#         self.queries = queries

#     def run(self, pipeline, module_ids):
#         result = set()
#         for query in queries:
#             result = result.union(query.run(pipeline, module_ids))

################################################################################
from __future__ import division

import xml.sax.saxutils

from vistrails.core.utils import memo_method

# convenience method that does the full html extract if PyQt is loaded
def extract_text(escaped_html_str):
    notes = xml.sax.saxutils.unescape(escaped_html_str)
    try:
        from PyQt4 import QtGui
    except ImportError:
        return str(notes)
    else:
        fragment = QtGui.QTextDocumentFragment.fromHtml(notes)
        return str(fragment.toPlainText())

# The queries are old and are preserved for reference.  Some code is
# quite old (references to vis_application, for example).

class Query(object):

    def upstream(self, graph, id):
        return graph.bfs(id).keys()

    weekdayMap = {'mo':0, 'tu':1, 'we':2, 'th':3, 'fr':4, 'sa':5, 'su':6}

    def weekday(self, time, weekday):
        return time.weekday() == self.weekdayMap[weekday.lower()[:2]]

    def computeIndices(self):
        self.versionDict = {}

        if not len(self.queryResult):
            return
        elif len(self.queryResult[0]) == 1:
            for (v,) in self.queryResult:
                self.versionDict[v] = {}
        elif len(self.queryResult[0]) == 2:
            for (v, m) in self.queryResult:
                if not self.versionDict.has_key(v):
                    self.versionDict[v] = {m: []}
                self.versionDict[v][m] = {}
        else:
            assert len(self.queryResult[0]) == 3
            for (v, m, e) in self.queryResult:
                if not self.versionDict.has_key(v):
                    self.versionDict[v] = {m: [e]}
                elif not self.versionDict[v].has_key(m):
                    self.versionDict[v][m] = [e]
                else:
                    self.versionDict[v][m].append(e)

    def match(self, vistrail, action):
        return action.timestep in self.versionDict

    def matchModule(self, version_id, module):
        return (self.tupleLength < 2 or
                (version_id in self.versionDict and
                 module.id in self.versionDict[version_id]))

    def executionInstances(self, version_id, module_id):
        versionDict = self.versionDict
        assert self.tupleLength == 3
        if versionDict[version_id].has_key(module_id):
            return versionDict[version_id][module_id]
        else:
            return None

class Query1a(Query):

    def run(self, vistrail, name):
        result = []
        versions = vistrail.get_tagMap().itervalues()
        for version in versions:
            p = vistrail.getPipeline(version)
            for module_id, module in p.modules.iteritems():
                if module.name == 'FileSink':
                    for f in module.functions:
                        if (f.name == 'outputName' and
                            len(f.params) == 1 and
                            f.params[0].value() == 'atlas-x.gif'):
                            result.append((version, module_id))
        self.queryResult = result
        self.tupleLength = 2
        self.computeIndices()
        return result


class Query1b(Query):

    def run(self, vistrail, name):
        result = []
        versions = vistrail.get_tagMap().itervalues()
        for version in versions:
            p = vistrail.getPipeline(version)
            ms = []
            for module_id, module in p.modules.iteritems():
                if module.name == 'FileSink':
                    for f in module.functions:
                        if (f.name == 'outputName' and
                            len(f.params) == 1 and
                            f.params[0].value() == 'atlas-x.gif'):
                            ms.append(module_id)
            s = set()
            inv_graph = p.graph.inverse()
            for m in ms:
                s = s.union(set(inv_graph.bfs(m).keys() + [m]))
            for m in s:
                result.append((version, m))
        self.queryResult = result
        self.tupleLength = 2
        self.computeIndices()
        return result


class Query1c(Query):

    def run(self, vistrail, name):
        import vistrails.gui.vis_application
        c = vistrails.gui.vis_application.logger.db.cursor()
        c.execute("""
        select distinct module_id, wf_version from
        wf_exec, exec, vistrails
        where
           wf_exec.wf_exec_id = exec.wf_exec_id and
           vistrails.vistrails_name = %s and
           vistrails.vistrails_id = wf_exec.vistrails_id""", name)
        lst = list(c.fetchall())
        versions = set(x[1] for x in lst)
        executions = set(lst)
        result = []
        for version in versions:
            p = vistrail.getPipeline(int(version))
            ms = []
            for module_id, module in p.modules.iteritems():
                if (module_id, version) not in executions:
                    continue
                if module.name == 'FileSink':
                    for f in module.functions:
                        if (f.name == 'outputName' and
                            len(f.params) == 1 and
                            f.params[0].value() == 'atlas-x.gif'):
                            ms.append(module_id)
            s = set()
            inv_graph = p.graph.inverse()
            for m in ms:
                s = s.union(set(inv_graph.bfs(m).keys() + [m]))
            for m in s:
                result.append((int(version), m))
        self.queryResult = result
        self.tupleLength = 2
        self.computeIndices()
        return result
    

class Query3(Query):

    def run(self, vistrail, name):
        import vistrails.gui.vis_application
        c = vistrails.gui.vis_application.logger.db.cursor()
        c.execute("""
        select distinct module_id, wf_version from
        wf_exec, exec, vistrails
        where
           wf_exec.wf_exec_id = exec.wf_exec_id and
           vistrails.vistrails_name = %s and
           vistrails.vistrails_id = wf_exec.vistrails_id""", name)
        lst = list(c.fetchall())
        versions = set(x[1] for x in lst)
        executions = set(lst)
        result = []
        for version in versions:
            p = vistrail.getPipeline(int(version))
            ms = []
            for module_id, module in p.modules.iteritems():
                if (module_id, version) not in executions:
                    continue
                if module.name == 'FileSink':
                    for f in module.functions:
                        if (f.name == 'outputName' and
                            len(f.params) == 1 and
                            f.params[0].value() == 'atlas-x.gif'):
                            ms.append(module_id)
            s = set()
            inv_graph = p.graph.inverse()
            for m in ms:
                s = s.union(set(inv_graph.bfs(m).keys() + [m]))
            for m in s:
                if (p.modules[m].has_annotation_with_key('stage') and
                    p.modules[m].get_annotation_by_key('stage').value in \
                        ['3','4','5']):
                    result.append((int(version), m))
        self.queryResult = result
        self.tupleLength = 2
        self.computeIndices()
        return result
        
            
class Query4(Query):

    def run(self, vistrail, name):
        import vistrails.gui.vis_application
        c = vistrails.gui.vis_application.logger.db.cursor()
        c.execute("""
        select distinct exec.ts_start, exec.ts_end, exec_id, module_id, wf_version from
        wf_exec, exec, vistrails
        where
           exec.module_name = %s and
           wf_exec.wf_exec_id = exec.wf_exec_id and
           vistrails.vistrails_name = %s and
           vistrails.vistrails_id = wf_exec.vistrails_id""", ('AlignWarp', name))

        result = []
        for ts_start, ts_end, exec_id, module_id, wf_version in c.fetchall():
            p = vistrail.getPipeline(int(wf_version))
            m = p.modules[module_id]
            assert m.name == 'AlignWarp'
            # We assume here that no module takes longer than a day to execute.
            wd = 'monday'
            if self.weekday(ts_start, wd) or self.weekday(ts_end, wd):
                result.append((wf_version, module_id, exec_id))

        self.queryResult = result
        self.tupleLength = 3
        self.computeIndices()
        return result

class Query5(Query):

    @memo_method
    def pipeline(self, vistrail, version):
        return vistrail.getPipeline(version)

    @memo_method
    def upstream_set(self, graph, m_id):
        return set(self.upstream(graph, m_id))

    def run(self, vistrail, name):
        import vistrails.gui.vis_application
        c = vistrails.gui.vis_application.logger.db.cursor()
        c.execute("""
        select distinct
           wf_exec.wf_version, exec.module_id
        from
           wf_exec, exec, annotation, vistrails
        where
           wf_exec.wf_exec_id = exec.wf_exec_id and
           vistrails.vistrails_name = %s and
           vistrails.vistrails_id = wf_exec.vistrails_id and
           annotation.exec_id = exec.exec_id and
           annotation.key = %s and
           annotation.value = %s""", (name, 'global maximum', '4095'))
        result = []
        all = c.fetchall()
        #print len(all)
        version_module = {}
        for wf_version, module_id in all:
            if not wf_version in version_module:
                version_module[wf_version] = set([module_id])
            else:
                version_module[wf_version].add(module_id)
        #print vistrail.get_tagMap()
        for wf_version, module_ids in version_module.iteritems():
            #print wf_version
            p = self.pipeline(vistrail, int(wf_version))
            inv_graph = p.graph.inverse()
            s = set()
            for m_id, module in p.modules.iteritems():
                if module.name == 'FileSink':
                    for f in module.functions:
                        if (f.name == 'outputName' and
                            len(f.params) == 1 and
                            f.params[0].value().find('atlas') != -1):
                            u_id = self.upstream_set(inv_graph, m_id)
                            if module_ids.intersection(u_id):
                                s = s.union(u_id).union(set([m_id]))
                            break
            for m in s:
                result.append((int(wf_version), m))
        self.queryResult = result
        self.tupleLength = 2
        self.computeIndices()
        return result
        

class Query6(Query):

    def run(self, vistrail, name):
        import vistrails.gui.vis_application
        c = vistrails.gui.vis_application.logger.db.cursor()
        c.execute("""
        select distinct module_id, wf_version from
        wf_exec, exec, vistrails
        where
           exec.module_name = %s and
           wf_exec.wf_exec_id = exec.wf_exec_id and
           vistrails.vistrails_name = %s and
           vistrails.vistrails_id = wf_exec.vistrails_id""", ('SoftMean', name))
        result = []
        for (module_id, wf_version) in c.fetchall():
            p = vistrail.getPipeline(int(wf_version))
            m = p.modules[module_id]
            assert m.name == 'SoftMean'
            inv_graph = p.graph.inverse()
            up = self.upstream(inv_graph, module_id)
            up.append(module_id)
            found = False
            for up_id in up:
                up_module = p.modules[up_id]
                if up_module.name == 'AlignWarp':
                    for f in up_module.functions:
                        if (f.name == 'model' and
                            len(f.params) == 1 and
                            f.params[0].value() == 12):
                            found = True
                            break
                if found:
                    break
            if found:
                for up_id in up:
                    result.append((wf_version, up_id))
        self.queryResult = result
        self.tupleLength = 2
        self.computeIndices()
        return result


class Query8(Query):

    def run(self, vistrail, name):
        result = []
        versions = vistrail.get_tagMap().itervalues()
        for version in versions:
            s = set()
            p = vistrail.getPipeline(version)
            inv_graph = p.graph.inverse()
            for module_id, module in p.modules.iteritems():
                if module.name == 'AlignWarp':
                    found = False
                    u_ids = self.upstream(inv_graph, module_id)
                    for i in u_ids:
                        if (p.modules[i].has_annotation_with_key('center')
                            and p.modules[i].get_annotation_by_key('center') \
                                == 'UChicago'):
                            found = True
                            break
                    if found:
                        s = s.union(set(u_ids + [module_id]))
            for m in s:
                result.append((version, m))
        self.queryResult = result
        self.tupleLength = 2
        self.computeIndices()
        return result


class Query9(Query):

    def run(self, vistrail, name):
        result = []
        versions = vistrail.get_tagMap().itervalues()
        for version in versions:
            s = set()
            p = vistrail.getPipeline(version)
            inv_graph = p.graph.inverse()
            for module_id, module in p.modules.iteritems():
                annot = module.annotations
                if (module.has_annotation_with_key('studyModality')
                    and module.get_annotation_by_key('studyModality') in \
                        ['visual', 'audio', 'speech']):
                    s = s.union(set(self.upstream(inv_graph, module_id) + [module_id]))
            for m in s:
                result.append((version, m))
        self.queryResult = result
        self.tupleLength = 2
        self.computeIndices()
        return result
